;;; **********************************************************************
;;; Copyright (C) 2005 Todd Ingals, <testcase (at) asu (dot) edu>
;;;
;;; This program is free software; you can redistribute it and/or
;;; modify it under the terms of the Lisp Lesser Gnu Public License.
;;; See http://www.cliki.net/LLGPL for the text of this agreement.
;;; **********************************************************************

;;; $Name: rel-2_0_1 $

(in-package :cm)

(defmethod recv ((io portmidi-stream) &key (resolution 5) (priority 20))
  (if (io-open io)
      (if (= (pm:recv-state) 0)
	  (let ((data (rt-stream-receive-data io )))
	    (if (= (pm:set-input (elt data 0)
				 (elt data 1)
				 (elt data 2)) 
		   0)
		(if (= (pm:recv-start priority resolution ) 0)
		    (format t "; recv running on stream ~s~%"
			    io))))
	  (warn "recv is already running on ~s" io))
      (warn "stream ~s is not open" io))
  (values))

(defmethod recv-stop ((io portmidi-stream))
  (if (io-open io)
      (if (= (pm:recv-state) 1)
	  (progn
	    (pm:recv-stop)
	    (format t "; recv stopped on stream ~s" io))
	  (warn "recv is not running on ~s" io)))
  (values))

(defmethod recv-set! ((io portmidi-stream) hook &key recv-mode)
  (if (io-open io)
      (if (functionp hook)
	  (progn
	    (setq pm::*pm-recv-callback* 
		  (lambda (m) 
		    (ignore-errors
		      (funcall hook (pm-message->midi-message m)))))
	    )
	  (progn
	    (setq pm::*pm-recv-callback* (lambda (x) x (values)))
	    nil))
      (warn "stream ~s is not open" io))
  (values))

(defmethod recv? ((io portmidi-stream) )
  (if (io-open io)
      (if (= (pm:recv-state) 0)
	  ':stopped
	  ':running)
      (warn "stream ~s is not open" io)))

;;; eof
